/*
**  pth_high.c -- PTH high-level replacement functions
**
**  Copyright (c) 1999 Ralf S. Engelschall <rse@engelschall.com>
**
**  This file is part of PTH, a non-preemptive thread scheduling library
**  which can be found at http://www.gnu.org/software/pth/.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
**  USA, or contact Ralf S. Engelschall <rse@engelschall.com>.
*/

/*
 *  These functions used by the applications instead of the
 *  regular Unix/POSIX functions. When the regular functions would
 *  block these variants let only the tread sleep.
 */

#include "pth_p.h"

/* PTH variant of fork(2) to fork a thread into a new process */
pid_t pth_fork(void)
{
    pid_t pid;

    pid = fork();
    if (pid != 0) {
        /* parent has nothing to do. It just returns 
           the child pid to the calling thread */
    }
    else {
        /* child has to kick out all threads except for
           the currently running thread and the scheduler */
        pth_scheduler_drop();
    }
    return pid;
}

/* PTH variant of usleep(3) */
int pth_usleep(unsigned int usec)
{
    pth_time_t until;
    pth_time_t offset;
    pth_event_t ev;
    static pth_key_t ev_key = PTH_KEY_INIT;

    /* calculate asleep time */
    offset = pth_time(0, usec);
    pth_time_set(&until, PTH_TIME_NOW);
    pth_time_add(&until, &offset);

    /* and let thread sleep until this time is elapsed */
    ev = pth_event(PTH_EVENT_TIME|PTH_MODE_STATIC, &ev_key, until);
    pth_wait(ev);

    return 0;
}

/* PTH variant of sleep(3) */
unsigned int pth_sleep(unsigned int sec)
{
    pth_time_t until;
    pth_time_t offset;
    pth_event_t ev;
    static pth_key_t ev_key = PTH_KEY_INIT;

    /* calculate asleep time */
    offset = pth_time(sec, 0);
    pth_time_set(&until, PTH_TIME_NOW);
    pth_time_add(&until, &offset);
    
    /* and let thread sleep until this time is elapsed */
    ev = pth_event(PTH_EVENT_TIME|PTH_MODE_STATIC, &ev_key, until);
    pth_wait(ev);

    return 0;
}

/* PTH variant of POSIX pthread_sigmask(3) */
int pth_sigmask(int how, const sigset_t *set, sigset_t *oset)
{
    /* that's really easy in PTH, we just call POSIX sigprocmask().
       Because the PTH machine context switching is written to
       safe and restore also the complete signal set */
    return sigprocmask(how, set, oset);
}

/* PTH variant of POSIX sigwait(3) */
int pth_sigwait(const sigset_t *set, int *sigp)
{
    return pth_sigwait_ev(set, sigp, NULL);
}

/* PTH variant of POSIX sigwait(3) with extra events */
int pth_sigwait_ev(const sigset_t *set, int *sigp, pth_event_t ev_extra)
{
    pth_event_t ev;
    static pth_key_t ev_key = PTH_KEY_INIT;
    sigset_t pending;
    int sig;

    if (set == NULL || sigp == NULL)
        return_errno(EINVAL, EINVAL);

    /* check whether signal is already pending */
    if (sigpending(&pending) < 0)
        sigemptyset(&pending);
    for (sig = 1; sig < PTH_NSIG; sig++) {
        if (sigismember(set, sig) && sigismember(&pending, sig)) {
            pth_util_sigdelete(sig);
            *sigp = sig;
            return 0;
        }
    }

    /* create event and wait on it */
    ev = pth_event(PTH_EVENT_SIGS|PTH_MODE_STATIC, &ev_key, set, sigp);
    if (ev_extra != NULL)
        pth_event_concat(ev, ev_extra, NULL);
    pth_wait(ev);
    if (ev_extra != NULL) {
        pth_event_isolate(ev);
        if (!pth_event_occurred(ev))
            return_errno(EINTR, EINTR);
    }

    /* nothing to do, scheduler has already set *sigp for us */
    return 0;
}

/* PTH variant of connect(2) */
int pth_connect(int s, struct sockaddr *addr, int addrlen)
{
    return pth_connect_ev(s, addr, addrlen, NULL);
}

/* PTH variant of connect(2) with extra events */
int pth_connect_ev(int s, struct sockaddr *addr, int addrlen, pth_event_t ev_extra)
{
    pth_event_t ev;
    static pth_key_t ev_key = PTH_KEY_INIT;
    int rv, err, errlen;
  
    /* try to connect */
    while ((rv = connect(s, addr, addrlen)) == -1 && errno == EINTR)
        ;
    /* when it is still on progress wait until socket is really writeable */
    if (rv == -1 && errno == EINPROGRESS) {
        ev = pth_event(PTH_EVENT_FD|PTH_UNTIL_FD_WRITEABLE|PTH_MODE_STATIC, &ev_key, s);
        if (ev_extra != NULL)
            pth_event_concat(ev, ev_extra, NULL);
        pth_wait(ev);
        if (ev_extra != NULL) {
            pth_event_isolate(ev);
            if (!pth_event_occurred(ev))
                return_errno(-1, EINTR);
        }
        errlen = sizeof(err);
        if (getsockopt(s, SOL_SOCKET, SO_ERROR, (void *)&err, &errlen) == -1)
            return -1;
        if (err == 0) 
            return 0;
        return_errno(rv, err);
    }
    return rv;
}

/* PTH variant of accept(2) */
int pth_accept(int s, struct sockaddr *addr, int *addrlen)
{
    return pth_accept_ev(s, addr, addrlen, NULL);
}

/* PTH variant of accept(2) with extra events */
int pth_accept_ev(int s, struct sockaddr *addr, int *addrlen, pth_event_t ev_extra)
{
    pth_event_t ev;
    static pth_key_t ev_key = PTH_KEY_INIT;
    int rv;

    /* poll socket via accept */
    ev = NULL;
    while ((rv = accept(s, addr, addrlen)) == -1
           && (errno == EAGAIN || errno == EWOULDBLOCK)) {
        /* do lazy event allocation */
        if (ev == NULL) {
            ev = pth_event(PTH_EVENT_FD|PTH_UNTIL_FD_READABLE|PTH_MODE_STATIC, &ev_key, s);
            if (ev_extra != NULL)
                pth_event_concat(ev, ev_extra, NULL);
        }
        /* wait until accept has a chance */
        pth_wait(ev);
        /* check for the extra events */
        if (ev_extra != NULL) {
            pth_event_isolate(ev);
            if (!pth_event_occurred(ev))
                return_errno(-1, EINTR);
        }
    }
    return rv;
}

/* PTH variant of read(2) */
ssize_t pth_read(int fd, void *buf, size_t nbytes)
{
    return pth_read_ev(fd, buf, nbytes, NULL);
}

/* PTH variant of read(2) with extra event(s) */
ssize_t pth_read_ev(int fd, void *buf, size_t nbytes, pth_event_t ev_extra)
{
    struct timeval delay;
    pth_event_t ev;
    static pth_key_t ev_key = PTH_KEY_INIT;
    fd_set fds;
    int n;

    /* first directly poll filedescriptor for readability 
       to avoid unneccessary (and resource consuming because of context
       switches, etc) event handling through the scheduler */
    FD_ZERO(&fds);
    FD_SET(fd, &fds);
    delay.tv_sec  = 0; 
    delay.tv_usec = 0;
    while ((n = select(fd+1, &fds, NULL, NULL, &delay)) < 0 
           && errno == EINTR) ; 

    /* when filedescriptor is still not readable, 
       let thread sleep until it is or event occurs */
    if (n < 1) {
        ev = pth_event(PTH_EVENT_FD|PTH_UNTIL_FD_READABLE|PTH_MODE_STATIC, &ev_key, fd);
        if (ev_extra != NULL)
            pth_event_concat(ev, ev_extra, NULL);
        n = pth_wait(ev);
        if (ev_extra != NULL) {
            pth_event_isolate(ev);
            if (!pth_event_occurred(ev))
                return_errno(-1, EINTR);
        }
    }

    /* now perform the actual read */
    while ((n = read(fd, buf, nbytes)) < 0 
           && errno == EINTR) ;
    return n;
}

/* PTH variant of write(2) */
ssize_t pth_write(int fd, const void *buf, size_t nbytes)
{
    return pth_write_ev(fd, buf, nbytes, NULL);
}

/* PTH variant of write(2) with extra event(s) */
ssize_t pth_write_ev(int fd, const void *buf, size_t nbytes, pth_event_t ev_extra)
{
    struct timeval delay;
    pth_event_t ev;
    static pth_key_t ev_key = PTH_KEY_INIT;
    fd_set fds;
    int n;

    /* first directly poll filedescriptor for writeability 
       to avoid unneccessary (and resource consuming because of context
       switches, etc) event handling through the scheduler */
    FD_ZERO(&fds);
    FD_SET(fd, &fds);
    delay.tv_sec  = 0; 
    delay.tv_usec = 0;
    while ((n = select(fd+1, NULL, &fds, NULL, &delay)) < 0 
           && errno == EINTR) ; 

    /* when filedescriptor is still not writeable, 
       let thread sleep until it is or event occurs */
    if (n < 1) {
        ev = pth_event(PTH_EVENT_FD|PTH_UNTIL_FD_WRITEABLE|PTH_MODE_STATIC, &ev_key, fd);
        if (ev_extra != NULL)
            pth_event_concat(ev, ev_extra, NULL);
        pth_wait(ev);
        if (ev_extra != NULL) {
            pth_event_isolate(ev);
            if (!pth_event_occurred(ev))
                return_errno(-1, EINTR);
        }
    }

    /* now perform the actual write */
    while ((n = write(fd, buf, nbytes)) < 0 
           && errno == EINTR) ;
    return n;
}

/* PTH variant of waitpid(2) */
pid_t pth_waitpid(pid_t wpid, int *status, int options)
{
    pth_event_t ev;
    static pth_key_t ev_key = PTH_KEY_INIT;
    pid_t pid;
    
    /* do a non-blocking poll for the pid */
    pid = waitpid(wpid, status, options | WNOHANG);
    /* if pid was found or caller requested a polling return immediately */
    if ((pid != 0) || (options & WNOHANG))
        return pid;

    /* else create a pid event and block current thread until it occurred */
    ev = pth_event(PTH_EVENT_PID|PTH_MODE_STATIC, 
                   &ev_key, wpid, status, options);
    pth_wait(ev);

    /* finally do the waitpid again for the result value and return it */
    pid = waitpid(wpid, status, options);
    return pid;
}

/*
 * implementation of a convinient greedy tread-safe line reading function to
 * avoid slow byte-wise reading from filedescriptors - which is important for
 * high-performance situations.
 */

#define READLINE_MAXLEN 1024
static pth_key_t  readline_key;
static pth_once_t readline_once_ctrl = PTH_ONCE_INIT;

typedef struct {
    int   rl_cnt;
    char *rl_bufptr;
    char  rl_buf[READLINE_MAXLEN];
} readline_buf;

static void readline_buf_destroy(void *vp)
{
    free(vp);
    return;
}

static void readline_init(void *vp)
{
    pth_key_create(&readline_key, readline_buf_destroy);
    return;
}

static ssize_t readline_read(readline_buf *rl, int fd, char *ptr, pth_event_t ev_extra)
{
    if (rl->rl_cnt <= 0) {
        for (;;) {
            if ((rl->rl_cnt = pth_read_ev(fd, rl->rl_buf, READLINE_MAXLEN, ev_extra)) < 0)
                return -1;
            else if (rl->rl_cnt == 0)
                return 0;
            rl->rl_bufptr = rl->rl_buf;
            break;
        }
    }
    rl->rl_cnt--;
    *ptr = *rl->rl_bufptr++;
    return 1;
}

ssize_t pth_readline(int fd, void *buf, size_t buflen)
{
    return pth_readline_ev(fd, buf, buflen, NULL);
}

ssize_t pth_readline_ev(int fd, void *buf, size_t buflen, pth_event_t ev_extra)
{
    int n, rc;
    char c, *cp;
    readline_buf *rl;

    pth_once(&readline_once_ctrl, readline_init, NULL);
    if ((rl = (readline_buf *)pth_key_getdata(readline_key)) == NULL) {
        rl = malloc(sizeof(readline_buf));
        rl->rl_cnt = 0;
        rl->rl_bufptr = NULL;
        pth_key_setdata(readline_key, rl);
    }
    cp = (char *)buf;
    for (n = 1; n < buflen; n++) {
        if ((rc = readline_read(rl, fd, &c, ev_extra)) == 1) {
            if (c == '\r') {
                n--;
                continue;
            }
            *cp++ = c;
            if (c == '\n')
                break;
        } 
        else if (rc == 0) {
            if (n == 1)
                return 0;
            else
                break;
        } 
        else
            return -1;
    }
    *cp = NUL;
    return n;
}

